/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: database.c 150 2004-05-24 23:52:35Z pasi $ */


#include <mysql.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>

#include "structs.h"
#include "spamserv.h"
#include "database.h"
#include "pseudoclient.h"
#include "cycles.h"
#include "irc.h"
#include "log.h"
#include "magic.h"
#include "util.h"

MYSQL *sql_c = NULL;

static  int db_open_connection(void);
static void db_close_connection(void);
static void db_error(const char *, ...);
static  int db_query(char *, ...);

int database_init(void)
{
	int error;
	
	s_log(LOG_DEBUG, "Initializing MySQL connection.");
	if ((error = db_open_connection()) < 0) {
		db_error("error %d", error);
		return -EDBERROR;
	}
	s_log(LOG_DEBUG, "Connected to MySQL");
	return 0;
}

void database_deinit(void)
{
	db_close_connection();
}

int database_load_channels(void)
{
	MYSQL_RES *result;
	MYSQL_ROW row;
	
	if (db_query("SELECT channel FROM channels") != 0) {
		db_error("database_load_channels(): %s",
			 mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	result = mysql_store_result(sql_c);
	if (result == NULL) {
		db_error("Could not fetch query result.");
		return -EDBERROR;
	}
	while ((row = mysql_fetch_row(result)) != NULL)
		(void)s_add_pseudochan(row[0]);
	mysql_free_result(result);
	return 0;
}

int database_load_spamstrs(void)
{
        MYSQL_RES *result;
        MYSQL_ROW row;

	/* XXX we're not really using the reason field from here? */
        if (db_query("SELECT id, string, mask, added, flags, reason, "
	             "matches, lastmatch, comment FROM spamstrings") != 0) {
                db_error("database_load_channels(): %s",
                         mysql_error(sql_c));
                return -EDBQUERYFAIL;
        }
        result = mysql_store_result(sql_c);
        if (result == NULL) {
                db_error("Could not fetch query result.");
                return -EDBERROR;
        }
        while ((row = mysql_fetch_row(result)) != NULL)
		 spam_str_add(row[1], row[2], (time_t)atol(row[3]),
			     atoi(row[4]),  strtoul(row[0], NULL, 0),
			     strtoul(row[6], NULL, 0), (time_t)atol(row[7]),
			     TRUE, NULL);
	mysql_free_result(result);
        return 0;
}

int database_load_cycles(void)
{
	MYSQL_RES *result;
	MYSQL_ROW row;
	
	if (db_query("SELECT channel, delay FROM cycle_channels") != 0) {
		db_error("database_load_cycles(): %s",
		mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	result = mysql_store_result(sql_c);
	if (result == NULL) {
		db_error("Could not fetch query result.");
		return -EDBERROR;
	}
	while ((row = mysql_fetch_row(result)) != NULL) 
		cycle_chan_add(row[0], atoi(row[1]), TRUE);
	mysql_free_result(result);
	return 0;
}

int database_add_forbid(const char *channel, const char *addedby,
			const char *reason)
{
	if (db_query("INSERT INTO forbidden_channels "
		     "(channel, addedby, reason) VALUES "
		     "('%s', '%s', '%s')", channel, addedby, reason) != 0) {
		/* ER_DUP_ENTRY */
		if (mysql_errno(sql_c) == 1062)
			return -EDBDUPLICATE;
		db_error("database_add_forbid():% s", mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

int database_del_forbid(const char *channel)
{
	if (db_query("DELETE FROM forbidden_channels "
		     "WHERE channel = '%s'", channel) != 0) {
		db_error("database_del_forbid(): %s", mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

int database_is_forbid(const char *channel, char *reason)
{
	MYSQL_RES *result;
	MYSQL_ROW row;

	if (db_query("SELECT reason FROM forbidden_channels "
		     "WHERE channel = '%s'", channel) != 0) {
		db_error("database_is_forbid(): %s", mysql_error(sql_c));
		/* this differs from functionality elsewhere */
		return 0;
	}
	result = mysql_store_result(sql_c);
	if (!result)
		return 0;
	row = mysql_fetch_row(result);
	if (row == NULL) {
		return 0;
	}
	g_strlcpy(reason, row[0], 513);
	mysql_free_result(result);
	return 1;
}

int database_add_channel(const char *channel, const char *addedby, time_t time)
{
	if (db_query("INSERT INTO channels (channel, mask, invitetime) "
		     "VALUES ('%s', '%s', '%l')",  channel, addedby,
		     time) != 0) {
		db_error("database_add_channel(): %s",
			 mysql_error(sql_c));
		return  -EDBQUERYFAIL;
	}
	return 0;
}

int database_del_channel(const char *channel)
{
	if (db_query("DELETE FROM channels WHERE channel='%s'",
		     channel) != 0) {
		db_error("database_del_channel(): %s",
			 mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

int database_add_spamstr(struct spam_str *str)
{
        if (db_query("INSERT INTO spamstrings (string, mask, added, "
		     "flags) VALUES ('%s', '%s', '%l', '%d')",
		     str->match_str->str, str->addedby,
		     str->added, str->flags) != 0) {
                db_error("database_add_spamstr(): %s",
                         mysql_error(sql_c));
                return  -EDBQUERYFAIL;
        }
	str->id = (unsigned long)mysql_insert_id(sql_c);
        return 0;
}

int database_add_cycle(const char *chan, const int delay)
{
	if (db_query("INSERT INTO cycle_channels (channel, delay) VALUES "
		     "('%s', '%d')", chan, delay) != 0) {
		db_error("database_add_cycle(): %s",
			 mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

int database_del_cycle(const char *chan)
{
	if (db_query("DELETE FROM cycle_channels WHERE channel='%s'",
		     chan) != 0) {
		db_error("database_del_cycle(): %s",
			 mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}


int database_update_reason(unsigned long id, const char *reason)
{
	if (db_query("UPDATE spamstrings SET reason='%s' "
		     "WHERE id = '%l'", reason, id) != 0) {
		db_error("database_update_reason(): %s",
		         mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

int database_update_comment(unsigned long id, const char *comment)
{
        if (db_query("UPDATE spamstrings SET comment='%s' "
                     "WHERE id = '%l'", comment, id) != 0) {
                db_error("database_update_comment(): %s",
                         mysql_error(sql_c));
                return -EDBQUERYFAIL;
        }
	return 0;
}

int database_update_matchdata(unsigned long id, unsigned long matches,
		              time_t lastmatch)
{
	if (db_query("UPDATE spamstrings SET lastmatch='%l', "
		     "matches='%l' WHERE id = '%l'", lastmatch,
		     matches, id) != 0) {
		db_error("database_update_matchdata(): %s",
			 mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

int database_del_spamstr(unsigned long id)
{
	if (db_query("DELETE from spamstrings WHERE id='%d'", id) != 0) {
		db_error("database_del_spamstr(): %s", mysql_error(sql_c));
		return  -EDBQUERYFAIL;
	}
	return 0;
}

int database_modify_flags(struct spam_str *str, int flags)
{
	if (db_query("UPDATE spamstrings SET flags='%d' "
		     "WHERE string = '%s'", flags,
		     str->match_str->str) != 0) {
		db_error("database_modify_flags(): %s",
		         mysql_error(sql_c));
		return -EDBQUERYFAIL;
	}
	return 0;
}

void database_list_channels_to(irc_user *user, const char *chan)
{
        MYSQL_RES *result;
        MYSQL_ROW row;

        if (db_query("SELECT channel,mask,invitetime FROM "
		     "channels WHERE channel LIKE '%s' ORDER BY channel",
		     chan) != 0) {
                db_error("database_list_channels(): %s",
                         mysql_error(sql_c));
                return;
        }
        result = mysql_store_result(sql_c);
        if (result == NULL) {
                db_error("Could not fetch query result.");
                return;
        }
        while ((row = mysql_fetch_row(result)) != NULL)
                i_send_notice(user, "%-30s (by %s at %s)", row[0], row[1],
			      s_timestamp_to_human((time_t)strtoul(row[2],
					           NULL, 0)));
	mysql_free_result(result);
        return;
}

void database_list_cycles_to(irc_user *user, const char *chan)
{
	MYSQL_RES *result;
	MYSQL_ROW row;

	if (db_query("SELECT channel, delay FROM cycle_channels "
		     "WHERE channel like '%s' ORDER by channel",
		     chan) != 0) {
		db_error("database_list_cycles_to(): %s",
			 mysql_error(sql_c));
		return;
	}
	result = mysql_store_result(sql_c);
	if (result == NULL) {
		db_error("Could not fetch query result.");
		return;
	}
	while ((row = mysql_fetch_row(result)) != NULL)
		i_send_notice(user, "%-30s (%s minute cycles)",
			      row[0], row[1]);
	mysql_free_result(result);
	return;
}

void database_list_forbids_to(irc_user *user, const char *chan)
{
	MYSQL_RES *result;
	MYSQL_ROW row;

	if (db_query("SELECT channel, reason, addedby FROM forbidden_channels"
		     " WHERE channel LIKE '%s' ORDER BY channel",
		     chan) != 0) {
		db_error("database_list_forbids_to(): %s", mysql_error(sql_c));
		return;
	}
	result = mysql_store_result(sql_c);
	if (result == NULL) {
		db_error("Could not fetch query result.");
		return;
	}
	while ((row = mysql_fetch_row(result)) != NULL)
		i_send_notice(user, "%s '%s' (%s)", row[0], row[1], row[2]);
	mysql_free_result(result);
	return;
}

static void db_error(const char *fmt, ...)
{
	va_list ap;
	char buf[2048];

	va_start(ap, fmt);
	g_vsnprintf(buf, sizeof(buf), fmt, ap);
	i_send_globops("database error! %s", buf);
	s_log(LOG_ERROR, "DB Error: %s", buf);
	va_end(ap);
	if (sql_c != NULL)
		db_close_connection();
	
}

/* parts of this taken directly from the k&r. and yes, i know %l behaves
 * in a weird fashion (actually uses unsigned %lu format, but hey.
 * works for us.) */
static int db_query(char *fmt, ...)
{
	GString *buf;
	int result;
	va_list ap;
	char *p, *sval;

	if (fmt == NULL)
		return -1;

	if (sql_c == NULL)
		if (db_open_connection() < 0)
			return -EDBNOCONN;

	buf = g_string_sized_new(64);
	va_start(ap, fmt);

	for (p = fmt; *p; p++) {
		if (*p != '%') {
			g_string_append_c(buf, *p);
			continue;
		}
		switch (*++p) {
		case 'd':
			g_string_append_printf(buf, "%d",
					       va_arg(ap, int));
			break;
		case 'f':
			g_string_append_printf(buf, "%f",
					       va_arg(ap, double));
			break;
		case 'l':
			g_string_append_printf(buf, "%lu",
					       va_arg(ap, long));
			break;
		case 's':
			for (sval = va_arg(ap, char *); *sval; sval++) {
				if (*sval == '\'' || *sval == '\"' ||
				    *sval == '\\')
					g_string_append_printf(buf, "\\%c", *sval);
				else
					g_string_append_c(buf, *sval);
			}
			break;
		default:
			g_string_append_c(buf, *p);
			break;
		}
	}
	va_end(ap);
	result = mysql_real_query(sql_c, buf->str, buf->len);
	g_string_free(buf, TRUE);
	if (result) {
		s_log(LOG_ERROR, "DB error: %s", mysql_error(sql_c));
	}
	return result;
}

static int db_open_connection(void)
{
	sql_c = mysql_init(NULL);
	if (sql_c == NULL)
		return -EDBINSUFFMEM;
	sql_c = mysql_real_connect(sql_c, conf.db.host, conf.db.user,
				   conf.db.pass, conf.db.database,
				   conf.db.port, NULL, 0);
	if (sql_c == NULL) {
		printf("ERROR: %s\n", mysql_error(sql_c));
		return -EDBCONNFAIL;
	}
	return 0;
}

static void db_close_connection(void)
{
	if (sql_c == NULL)
		return;
	mysql_close(sql_c);
	sql_c = NULL;
}
