/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: log.c 129 2004-05-22 01:43:15Z fj $ */

#include <stdio.h>
#include <stdarg.h>
#include <glib.h>
#include <assert.h>
#include <string.h>
#include "log.h"
#include "structs.h"

extern struct conf_t conf;

//#define DEBUG

const char *s_get_log_strid(lognums);

typedef struct t_logmethods {
	const char *filename;
	const char *strid;
	lognums loglevel;
	FILE *fp;
} logmethods;


logmethods loglevel[] = {
	{"debug",	"debug",	LOG_DEBUG,	NULL},
	{"autokill",	"akill",	LOG_AUTOKILL,	NULL},
	{"kill",	"kill",		LOG_KILL,	NULL},
	{"report",	"report",	LOG_REPORT,	NULL},
	{"misc",	"misc",		LOG_MISC,	NULL},
	{"fatal",	"fatal",	LOG_FATAL,	NULL},
	{"/dev/stdout",	"normal",	LOG_NORMAL,	NULL},
	{"/dev/stderr",	"error",	LOG_ERROR,	NULL},
	{"sent",	"sent",		LOG_SENT,	NULL},
	{"recv",	"recv",		LOG_RECV,	NULL},
	{"/dev/stdout",	"info",		LOG_INFO,	NULL},
	{NULL,		NULL,		0,		NULL}
};

const char *s_get_log_strid(lognums level)
{
	logmethods *lm;
	for (lm = loglevel; lm->strid != NULL; lm++) {
		if (lm->loglevel == level)
			return lm->strid;
	}
	return NULL;
}

const char *s_get_log_filename(lognums level)
{
	logmethods *lm;
	for (lm = loglevel; lm->strid != NULL; lm++) {
		if (lm->loglevel == level)
			return lm->filename;
	}
	return NULL;
}

FILE *s_get_log_fp(lognums level)
{
	logmethods *lm;
	char *filename;
	
	for (lm = loglevel; lm->filename != NULL; lm++) {
		if (lm->loglevel == level) {
			if (lm->fp == NULL) {	
				filename = (lm->filename[0] != '/') ? g_strdup_printf("%s/%s", conf.logpath, lm->filename) : g_strdup_printf("%s", lm->filename);
				lm->fp = fopen(filename, "a");
				g_free(filename);
			}
			return lm->fp;
		}
	}
	return NULL;
}

void s_log_close_handles()
{
	logmethods *lm;
	for (lm = loglevel; lm->filename != NULL; lm++) {
		if (lm->fp != NULL) {
			fclose(lm->fp);
			lm->fp = NULL;
		}
	}
}

void s_log(lognums level, const char *fmt, ...)
{
	va_list ap;
#ifndef DEBUG
	FILE *fp;
#endif
	char *buffer;
	

	if (strstr(conf.loglevels, s_get_log_strid(level)) == NULL)
			return;
	/* FIXME FIXME FIXME:: Very cheap way of checking. Don't do this. Evil. Evil. Evil. */
		
	
	va_start(ap, fmt);
	buffer = g_strdup_vprintf(fmt, ap);
	va_end(ap);

#ifdef DEBUG
	
	fprintf(stderr,"[%s] %s", s_get_log_strid(level), buffer);
	if (g_str_has_suffix(buffer, "\n") == FALSE)
		fputs("\n", stderr);
#else
	fp = s_get_log_fp(level);
        if (fp == NULL) {
		printf("[s_log error] unable to open file [%s]\n", s_get_log_filename(level));
		g_free(buffer);
		return;
	}
	
	fputs(buffer, fp);
	if (g_str_has_suffix(buffer, "\n") == FALSE)
		fputs("\n", fp);
	fflush(fp);
#endif
	g_free(buffer);
	
}
