	/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
	 *
	 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
	 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

	/* $Id: pseudoclient.c 153 2004-05-25 22:04:30Z pasi $ */

#include <glib.h>
#include <string.h>

#include "spamserv.h"
#include "structs.h"
#include "server.h"
#include "irc.h"
#include "pseudoclient.h"
#include "log.h"
#include "database.h"
#include "users.h"
#include "util.h"

#define FOREACH_PSEUDOCHAN(var, obj)					\
	for (var = 0;							\
	     var < (obj)->chanlistlen;					\
	     var++)

GSList *pseudoclients = NULL;

void s_introduce_pseudoclient(gpointer, gpointer);
void s_pseudoclient_add_channel(pclient *, char *);
void s_pseudoclient_send_sjoins(pclient *);
void s_free_pseudoclient(pclient *);
void s_free_pseudoclient_gpointer(gpointer, gpointer);
int s_pseudo_has_free_slots(gpointer, gpointer);
int s_del_pclient_chanlist_del(pclient *, char *);
static int s_pclient_comparechans(gpointer, gpointer);
pclient *s_find_pclient_with_channel(char *);
static int s_pseudocmp(gpointer, gpointer);


/* find a pseudoclient with free channel slots.
 * p1 = pseudoclient pointer
 * p2 = ignored 
 * this function now leaves 3 spots free in case users
 * want some specific client in their channel. 
 */
int s_pseudo_has_free_slots(gpointer p1, gpointer p2 __attr_unused__)
{
	unsigned int x;

	if (((pclient *)p1)->type & PCLIENT_MASTER)
		return -1;
	/* Just in case some wacko has set MAXPSEUDOCHANS to be
	 * smaller than 4 ... */
	x = (MAXPSEUDOCHANS < 4)  ? 0 : 3;
	return (((pclient *)p1)->freechannels < x) ? -1 : 0;
}

int s_invite_pseudoclient(char *channel, char *target, irc_user *user)
{
	pclient *p;
	char *mask;
	char s[512];

	if (channel == NULL)
		return -EGENERICERR;
	strip_all(channel);
	if (isvalid_chan(channel) == 0)
		return -EINVALIDCHAN;
	if (database_is_forbid(channel, s)) {
		i_send_notice(user, "Error: %s may not have a spamserv "
				"client: %s", channel, s);
		return -EGENERICERR;
	}

	if ((p = s_find_pclient_with_channel(channel)) != NULL) {
		i_send_notice(user, "There already is a "
				"SpamServ client [%s] in %s.",
				p->user.nick, channel);
		return -EGENERICERR;
	}
	s_fill_user_by_nick(user->nick, user);
	mask = g_strdup_printf("%s!%s@%s", user->nick, user->username,
			user->host);
	if (database_add_channel(channel, mask, time(NULL)) < 0) {
		g_free(mask);
		return -EDBERROR;
	}
	g_free(mask);
	if (g_ascii_strcasecmp(target, "SpamServ") != 0)
		p = s_add_pseudochan_target(channel, target);
	if (p == NULL) 
		p = s_add_pseudochan(channel);
	i_send_sjoin(&p->user, channel);
	send_raw(":%s NOTICE @%s :"
			"%s has invited a SpamServ bot (%s) into %s. "
			"It will report or disconnect from the network "
			"most spambots if they spam in %s. If you notice "
			"it fails to detect a spammer, please inform an "
			"IRC Operator. To remove it from %s, just /kick %s",
			p->user.nick, channel, user->nick, p->user.nick,
			channel, channel, channel, p->user.nick);
	return 0;
}

/* Returns NULL if the pseudoclient doesn't exist or it has no
 * free space for an extra channel. */
pclient *s_add_pseudochan_target(char *channel, char *pseudonick) {
	pclient* p;

	p = s_find_pseudoclient(pseudonick);
	if (p != NULL && p->freechannels > 0)
		s_pseudoclient_add_channel(p, channel);
	else
		p = NULL;
	return p;
}

/* adds channel to pseudoclient system and spawns pseudoclients
 * as necessary */
pclient *s_add_pseudochan(char *channel)
{
	GSList *tmp;
	pclient *p;

	s_log(LOG_DEBUG, "adding pseudochannel %s", channel);
	tmp = g_slist_find_custom(pseudoclients, NULL,
			(GCompareFunc)s_pseudo_has_free_slots);

	if (tmp == NULL) {
		p = s_add_pseudoclient();
		s_introduce_pseudoclient(p, NULL);
		s_pseudoclient_add_channel(p, channel);
	} else {
		p = tmp->data;
		s_pseudoclient_add_channel(p, channel);
	}
	return p;
}

void s_list_pseudochans_to(irc_user *user, char *mask)
{
	g_strdelimit(mask, "*?", '%');
	database_list_channels_to(user, mask);
	i_send_notice(user, "End of CHANNEL List");
}

/* should this be changed to return int and error if there's a
 * problem? */
void s_pseudoclient_add_channel(pclient *p, char *channel)
{
	if (p->freechannels) {
		p->channels[p->chanlistlen] = g_strndup(channel, MAXCHANLEN);
		p->chanlistlen++;
		p->freechannels--;
	}
}

void s_pseudoclient_parse_whois(irc_user *whoisser, char *nick)
{
	pclient *p;
	unsigned int i;
	GString *chanlist;

	p = s_find_pseudoclient(nick);
	if (p == NULL)
		return;
	chanlist = g_string_sized_new(30);
	FOREACH_PSEUDOCHAN(i, p) {
		g_string_append_printf(chanlist, "%s ", p->channels[i]);	
	}
	i_send_whois(whoisser, &p->user, chanlist);
	g_string_free(chanlist, TRUE);
}

/* send SJOIN for each channel stored for pseudoclient pointed by p */
void s_pseudoclient_send_sjoins(pclient *p)
{
	unsigned int i;
	char *chan;

	FOREACH_PSEUDOCHAN(i, p) {
		chan = p->channels[i];
		i_send_sjoin(&p->user, chan);
	}
}

/* add a special pseudoclient (usually SpamServ) for controlling the
 * system. */
void s_add_controlbot(void)
{
	pclient *p;

	p = g_malloc(sizeof(pclient));
	g_strlcpy(p->user.nick, "SpamServ", MAXNICKLEN);
	g_strlcpy(p->user.username, "spamserv", MAXUSERLEN);
	g_strlcpy(p->user.gcos, "Spam Fighter!", MAXREALLEN);
	g_strlcpy(p->user.host, "stop.spam", MAXHOSTLEN);
	p->type = PCLIENT_MASTER;
	p->freechannels = 2;
	p->chanlistlen = 0;
	s_log(LOG_INFO, "Starting spamserv control client.");
	pseudoclients = g_slist_append(pseudoclients, p);
	s_controlbot_add_chans(FALSE);
	status.stats.pseudoclient_count++;
}

void s_controlbot_add_chans(gboolean rehash)
{
	pclient *p;

	if ((p = s_find_pseudoclient("SpamServ")) == NULL)
		return;
	s_pseudoclient_add_channel(p, conf.infochan);
	if (g_ascii_strncasecmp(conf.infochan, conf.reportchan, MAXCHANLEN) != 0)
		s_pseudoclient_add_channel(p, conf.reportchan);
	if (rehash == TRUE)
		s_pseudoclient_send_sjoins(p);
}

void s_controlbot_del_chans(void)
{
	pclient *p;

	if ((p = s_find_pseudoclient("SpamServ")) == NULL)
		return;
	s_del_pclient_chanlist_del(p, conf.infochan);
	if (g_ascii_strncasecmp(conf.infochan, conf.reportchan, MAXCHANLEN) != 0)
		s_del_pclient_chanlist_del(p, conf.reportchan);
}

/* Allocate and prepare a new pseudoclient */
pclient *s_add_pseudoclient(void)
{
	pclient *p;

	p = g_malloc(sizeof(pclient));
	/* I don't think that this is too expensive. Chances
	 * that pseudoclient nicks are in use aren't that
	 * big anyway. 
	 * XXX: Add a warning perhaps? */
	/* XXX: Eerr this is already done elsewhere :)
	 *  (see introduce bit) */
	g_strlcpy(p->user.nick, getnick(), MAXNICKLEN);
	g_strlcpy(p->user.username, p->user.nick, MAXUSERLEN);
	g_strlcpy(p->user.gcos, "Spam Fighter!", MAXREALLEN);
	g_strlcpy(p->user.host, s_print_random_bothost(), MAXHOSTLEN);

	/* default the type to normal pseudoclient */
	p->type = PCLIENT_REGULAR;
	p->freechannels = MAXPSEUDOCHANS;
	p->chanlistlen = 0;

	s_log(LOG_DEBUG, "Adding pseudoclient %s.", p->user.nick);
	pseudoclients = g_slist_append(pseudoclients, p);
	status.stats.pseudoclient_count++;
	return p;
}

pclient *s_add_cycle_pseudo(pclient *p)
{
	p->type = PCLIENT_CYCLER | PCLIENT_INTRODUCED;
	pseudoclients = g_slist_append(pseudoclients, p);
	status.stats.pseudoclient_count++;
	return p;
}

/* send NICK and required SJOINs for a pseudoclient,
 * data = pclient *
 * args = ignored
 */
void s_introduce_pseudoclient(gpointer data, gpointer ignored __attr_unused__)
{
	pclient *p;

	/* the code to check for already introduced pseudos is here to
	 * allow us freely use s_introduce_pseudoclients(). this way
	 * we won't get nick collisions or anything. */
	if (!status.connected)
		return;
	p = (pclient *)data;
	if (p->type & PCLIENT_INTRODUCED) {
		s_log(LOG_DEBUG, "Attempted to introduce known pseudoclient.");
		return;
	}
	/* TODO: optimize. Right now if the user list contains the
	 * nick, we just pick a new one. This may result in several
	 * iterations through the list. */
	while (s_find_user(p->user.nick))
		g_strlcpy(p->user.nick, getnick(), MAXNICKLEN);

	i_send_nick(&p->user);
	s_pseudoclient_send_sjoins(p);
	p->type |= PCLIENT_INTRODUCED;
}

/* introduces all known spamserv clients */
void s_introduce_pseudoclients(void)
{
	if (!status.connected)
		return;
	g_slist_foreach(pseudoclients, (GFunc)s_introduce_pseudoclient,
			NULL);
}

/* frees a pseudoclient */
void s_free_pseudoclient(pclient *p)
{
	unsigned int i;

	s_log(LOG_DEBUG, "Freeing pseudoclient %s.", p->user.nick);
	if (status.connected)
		i_send_quit(&p->user, "SpamServ client exiting.");
	FOREACH_PSEUDOCHAN(i, p) {
		g_free(p->channels[i]);
	}
	g_free(p);
}

/* for use with s_free_pseudoclients(), calls
 * s_free_pseudoclient(). 
 * data = pointer to pclient 
 * args = ignored
 * PCLIENT_CYCLER bot is not free()ed because we never allocated
 * memory for it. 
 */
void s_free_pseudoclient_gpointer(gpointer data, gpointer args __attr_unused__)
{
	if (((pclient *)data)->type & PCLIENT_CYCLER)
		return;
	s_free_pseudoclient((pclient *)data);
}

/* free all pseudoclients */
void s_free_pseudoclients(void)
{
	g_slist_foreach(pseudoclients, (GFunc)s_free_pseudoclient, NULL);
	g_slist_free(pseudoclients);
	pseudoclients = NULL;
}

/* set introduced bit to 0 and make sure this pclient is ready for
 * new introducion later
 * data = pclient *
 * args = ignored */
void s_clean_pseudoclient(gpointer data, gpointer args __attr_unused__) {
	((pclient *)data)->type &= ~PCLIENT_INTRODUCED;
}

/*
 * IMPORTANT:
 *
 * called when a connection to server is lost. supposed to do
 * necessary cleanup for pseudoclient system, such as resetting
 * the introduced bits
 */
void s_pseudoclient_cleanup(void) {
	g_slist_foreach(pseudoclients, (GFunc)s_clean_pseudoclient, NULL);
}

/* FYI, i don't like this code :) */
int s_del_pclient_chanlist_del(pclient *p, char *c)
{
	unsigned int i;
	int found;

	found = 0;
	/* ok the logic here..
	 * we go through the channel list until we find the channel
	 * we're deleting. we free the data, and set a mark that
	 * we found it.
	 *
	 * we continue walking the list, and move all the following
	 * items back one step, and we stop before the end of the
	 * list. */
	FOREACH_PSEUDOCHAN(i, p) {
		if (g_ascii_strncasecmp(p->channels[i], c, MAXCHANLEN) == 0) {
			found++;
			i_send_part(&p->user, p->channels[i]);
			g_free(p->channels[i]);
			p->channels[i] = NULL;
		}
		if (found && i < p->chanlistlen) {
			p->channels[i] = p->channels[i + 1];
		}
	}
	/* decrease the counters */
	if (found) {
		p->chanlistlen--;
		p->freechannels++;
		return 0;
	} else
		return -1;
}

/*
 * called on kick, tihs doesn't send out error messages but just 
 * returns because we use this to check if the kick was really
 * targeted at one of our bots 
 */
void s_del_pseudochan_by_kick(char *nick, char *channel)
{
	pclient *p;

	if (nick == NULL || channel == NULL)
		return;
	if ((p = s_find_pseudoclient(nick)) == NULL)
		return;
	if (database_del_channel(channel) < 0)
		return;
	(void)s_del_pclient_chanlist_del(p, channel);
}

/* this is what should be called when a channel is removed from
 * pseudoclient channel database. */
int s_del_pseudochan(char *channel)
{
	pclient *p;

	if (database_del_channel(channel) < 0)
		return -ECHANDELFAIL;
	p = s_find_pclient_with_channel(channel);
	if (p == NULL)
		return -ECHANDELFAIL;
	if (s_del_pclient_chanlist_del(p, channel) == -1)
		return -ECHANDELFAIL;
	return 0;
}

char *s_pseudo_at_chan(char *channel)
{
	pclient *p;

	p = s_find_pclient_with_channel(channel);
	if (p == NULL)
		return NULL;
	else
		return p->user.nick;
}

/* 
 * for s_find_pclient_with_channel,
 * p1 = pclient *
 * p2 = char *
 */
static int s_pclient_comparechans(gpointer p1, gpointer p2)
{
	unsigned int i;
	pclient *p;

	p = (pclient *)p1;
	FOREACH_PSEUDOCHAN(i, p) {
		if (p->type & PCLIENT_MASTER)
			continue;
		if (g_ascii_strncasecmp(p->channels[i], p2, MAXCHANLEN) == 0)
			return 0;
	}
	return -1;
}

/*
 * find a pseudoclient inhabiting channel
 */
pclient *s_find_pclient_with_channel(char *chan)
{
	GSList *tmp;
	
	tmp = g_slist_find_custom(pseudoclients, chan,
			(GCompareFunc)s_pclient_comparechans);
	if (tmp == NULL)
		return NULL;
	else
		return tmp->data;
}

static int s_pseudocmp(gpointer p1, gpointer p2)
{
	return g_ascii_strncasecmp(((pclient *)p1)->user.nick, (char *)p2,
				  MAXNICKLEN);
}

pclient *s_find_pseudoclient(char *nick)
{
	GSList *tmp;

	tmp = g_slist_find_custom(pseudoclients, nick,
			          (GCompareFunc)s_pseudocmp);
	if (tmp == NULL)
		return NULL;
	else
		return tmp->data;
}

int s_pseudochan_forbid(char *line, char *mask)
{
	char *s;

	g_return_val_if_fail(line != NULL, -EGENERICERR);
	g_return_val_if_fail(mask != NULL, -EGENERICERR);
	
	if ((s = strchr(line, ' ')) == NULL ||
	     *(s + 1) == '\0')
		return -EINVALIDCMDFORMAT;
	*s = '\0'; s++;
	strip_all(line);
	if (!isvalid_chan(line))
		return -EINVALIDCHAN;
	return database_add_forbid(line, mask, s);
}

int s_pseudochan_unforbid(char *chan)
{
	g_return_val_if_fail(chan != NULL, -EGENERICERR);
	
	if (!isvalid_chan(chan))
		return -EINVALIDCHAN;
	return database_del_forbid(chan);
}

void s_pseudochan_forbids_list_to(irc_user *user, char *mask)
{
	g_return_if_fail(user != NULL);
	g_return_if_fail(mask != NULL);
	
	g_strdelimit(mask, "*?", '%');
	printf("%s hehe\n", mask);
	database_list_forbids_to(user, mask);
	i_send_notice(user, "End of list.");
}
